@@ -25,9 +25,10 @@
 #include "sysdeps.h"
 #include "gstvaapicompat.h"
 #include "gstvaapiutils.h"
-#include "gstvaapisurface.h"
-#include "gstvaapisubpicture.h"
+#include "gstvaapibufferproxy.h"
 #include "gstvaapifilter.h"
+#include "gstvaapisubpicture.h"
+#include "gstvaapisurface.h"
 #include <stdio.h>
 #include <stdarg.h>
 
@@ -164,12 +165,21 @@ vaapi_unmap_buffer (VADisplay dpy, VABufferID buf_id, gpointer * pbuf)
 gboolean
 vaapi_create_buffer (VADisplay dpy, VAContextID ctx, int type, guint size,
     gconstpointer buf, VABufferID * buf_id_ptr, gpointer * mapped_data)
+{
+  return vaapi_create_n_elements_buffer (dpy, ctx, type, size, buf, buf_id_ptr,
+      mapped_data, 1);
+}
+
+gboolean
+vaapi_create_n_elements_buffer (VADisplay dpy, VAContextID ctx, int type,
+    guint size, gconstpointer buf, VABufferID * buf_id_ptr,
+    gpointer * mapped_data, int num_elements)
 {
   VABufferID buf_id;
   VAStatus status;
   gpointer data = (gpointer) buf;
 
-  status = vaCreateBuffer (dpy, ctx, type, size, 1, data, &buf_id);
+  status = vaCreateBuffer (dpy, ctx, type, size, num_elements, data, &buf_id);
   if (!vaapi_check_status (status, "vaCreateBuffer()"))
     return FALSE;
 
@@ -227,6 +237,13 @@ string_of_VAProfile (VAProfile profile)
 #if VA_CHECK_VERSION(1,2,0)
       MAP (HEVCMain422_10);
       MAP (HEVCMain444);
+      MAP (HEVCMain444_10);
+      MAP (HEVCSccMain);
+      MAP (HEVCSccMain10);
+      MAP (HEVCSccMain444);
+#endif
+#if VA_CHECK_VERSION(1,8,0)
+      MAP (HEVCSccMain444_10);
 #endif
       MAP (HEVCMain);
       MAP (HEVCMain10);
@@ -257,6 +274,15 @@ string_of_VAEntrypoint (VAEntrypoint entrypoint)
       MAP (IDCT);
       MAP (MoComp);
       MAP (Deblocking);
+      MAP (EncSlice);
+      MAP (EncPicture);
+#if VA_CHECK_VERSION(0,39,1)
+      MAP (EncSliceLP);
+#endif
+      MAP (VideoProc);
+#if VA_CHECK_VERSION(1,0,0)
+      MAP (FEI);
+#endif
 #undef MAP
     default:
       break;
@@ -328,19 +354,27 @@ string_of_VARateControl (guint rate_control)
   switch (rate_control) {
     case VA_RC_NONE:
       return "None";
-#ifdef VA_RC_CQP
     case VA_RC_CQP:
       return "CQP";
-#endif
     case VA_RC_CBR:
       return "CBR";
     case VA_RC_VCM:
       return "VCM";
     case VA_RC_VBR:
       return "VBR";
-#ifdef VA_RC_VBR_CONSTRAINED
     case VA_RC_VBR_CONSTRAINED:
       return "VBR-Constrained";
+#if VA_CHECK_VERSION(0,39,1)
+    case VA_RC_MB:
+      return "MB";
+#endif
+#if VA_CHECK_VERSION(1,1,0)
+    case VA_RC_ICQ:
+      return "VA_RC_ICQ";
+#endif
+#if VA_CHECK_VERSION(1,3,0)
+    case VA_RC_QVBR:
+      return "VA_RC_QVBR";
 #endif
     default:
       break;
@@ -360,61 +394,39 @@ string_of_VARateControl (guint rate_control)
 guint
 to_GstVaapiChromaType (guint va_rt_format)
 {
-  guint chroma_type;
-
-  switch (va_rt_format) {
-    case VA_RT_FORMAT_YUV420:
-      chroma_type = GST_VAAPI_CHROMA_TYPE_YUV420;
-      break;
-    case VA_RT_FORMAT_YUV422:
-      chroma_type = GST_VAAPI_CHROMA_TYPE_YUV422;
-      break;
-    case VA_RT_FORMAT_YUV444:
-      chroma_type = GST_VAAPI_CHROMA_TYPE_YUV444;
-      break;
-    case VA_RT_FORMAT_YUV411:
-      chroma_type = GST_VAAPI_CHROMA_TYPE_YUV411;
-      break;
-    case VA_RT_FORMAT_YUV400:
-      chroma_type = GST_VAAPI_CHROMA_TYPE_YUV400;
-      break;
-    case VA_RT_FORMAT_RGB32:
-      chroma_type = GST_VAAPI_CHROMA_TYPE_RGB32;
-      break;
-    case VA_RT_FORMAT_RGB16:
-      chroma_type = GST_VAAPI_CHROMA_TYPE_RGB16;
-      break;
-    case VA_RT_FORMAT_RGBP:
-      chroma_type = GST_VAAPI_CHROMA_TYPE_RGBP;
-      break;
-    case VA_RT_FORMAT_YUV420_10BPP:
-      chroma_type = GST_VAAPI_CHROMA_TYPE_YUV420_10BPP;
-      break;
+  if (va_rt_format & VA_RT_FORMAT_YUV420)
+    return GST_VAAPI_CHROMA_TYPE_YUV420;
+  if (va_rt_format & VA_RT_FORMAT_YUV422)
+    return GST_VAAPI_CHROMA_TYPE_YUV422;
+  if (va_rt_format & VA_RT_FORMAT_YUV444)
+    return GST_VAAPI_CHROMA_TYPE_YUV444;
+  if (va_rt_format & VA_RT_FORMAT_YUV411)
+    return GST_VAAPI_CHROMA_TYPE_YUV411;
+  if (va_rt_format & VA_RT_FORMAT_YUV400)
+    return GST_VAAPI_CHROMA_TYPE_YUV400;
+  if (va_rt_format & VA_RT_FORMAT_RGB32)
+    return GST_VAAPI_CHROMA_TYPE_RGB32;
+  if (va_rt_format & VA_RT_FORMAT_RGB16)
+    return GST_VAAPI_CHROMA_TYPE_RGB16;
+  if (va_rt_format & VA_RT_FORMAT_RGBP)
+    return GST_VAAPI_CHROMA_TYPE_RGBP;
+  if (va_rt_format & VA_RT_FORMAT_YUV420_10BPP)
+    return GST_VAAPI_CHROMA_TYPE_YUV420_10BPP;
 #if VA_CHECK_VERSION(1,2,0)
-    case VA_RT_FORMAT_YUV422_10:
-      chroma_type = GST_VAAPI_CHROMA_TYPE_YUV422_10BPP;
-      break;
-    case VA_RT_FORMAT_YUV444_10:
-      chroma_type = GST_VAAPI_CHROMA_TYPE_YUV444_10BPP;
-      break;
-    case VA_RT_FORMAT_YUV420_12:
-      chroma_type = GST_VAAPI_CHROMA_TYPE_YUV420_12BPP;
-      break;
-    case VA_RT_FORMAT_YUV422_12:
-      chroma_type = GST_VAAPI_CHROMA_TYPE_YUV422_12BPP;
-      break;
-    case VA_RT_FORMAT_YUV444_12:
-      chroma_type = GST_VAAPI_CHROMA_TYPE_YUV444_12BPP;
-      break;
-    case VA_RT_FORMAT_RGB32_10:
-      chroma_type = GST_VAAPI_CHROMA_TYPE_RGB32_10BPP;
-      break;
+  if (va_rt_format & VA_RT_FORMAT_YUV422_10)
+    return GST_VAAPI_CHROMA_TYPE_YUV422_10BPP;
+  if (va_rt_format & VA_RT_FORMAT_YUV444_10)
+    return GST_VAAPI_CHROMA_TYPE_YUV444_10BPP;
+  if (va_rt_format & VA_RT_FORMAT_YUV420_12)
+    return GST_VAAPI_CHROMA_TYPE_YUV420_12BPP;
+  if (va_rt_format & VA_RT_FORMAT_YUV422_12)
+    return GST_VAAPI_CHROMA_TYPE_YUV422_12BPP;
+  if (va_rt_format & VA_RT_FORMAT_YUV444_12)
+    return GST_VAAPI_CHROMA_TYPE_YUV444_12BPP;
+  if (va_rt_format & VA_RT_FORMAT_RGB32_10)
+    return GST_VAAPI_CHROMA_TYPE_RGB32_10BPP;
 #endif
-    default:
-      chroma_type = 0;
-      break;
-  }
-  return chroma_type;
+  return 0;
 }
 
 /**
@@ -697,23 +709,27 @@ from_GstVaapiRateControl (guint value)
   switch (value) {
     case GST_VAAPI_RATECONTROL_NONE:
       return VA_RC_NONE;
-#ifdef VA_RC_CQP
     case GST_VAAPI_RATECONTROL_CQP:
       return VA_RC_CQP;
-#endif
     case GST_VAAPI_RATECONTROL_CBR:
       return VA_RC_CBR;
     case GST_VAAPI_RATECONTROL_VCM:
       return VA_RC_VCM;
     case GST_VAAPI_RATECONTROL_VBR:
       return VA_RC_VBR;
-#ifdef VA_RC_VBR_CONSTRAINED
     case GST_VAAPI_RATECONTROL_VBR_CONSTRAINED:
       return VA_RC_VBR_CONSTRAINED;
-#endif
-#ifdef VA_RC_MB
+#if VA_CHECK_VERSION(0,39,1)
     case GST_VAAPI_RATECONTROL_MB:
       return VA_RC_MB;
+#endif
+#if VA_CHECK_VERSION(1,1,0)
+    case GST_VAAPI_RATECONTROL_ICQ:
+      return VA_RC_ICQ;
+#endif
+#if VA_CHECK_VERSION(1,3,0)
+    case GST_VAAPI_RATECONTROL_QVBR:
+      return VA_RC_QVBR;
 #endif
   }
   GST_ERROR ("unsupported GstVaapiRateControl value %u", value);
@@ -726,24 +742,29 @@ to_GstVaapiRateControl (guint value)
   switch (value) {
     case VA_RC_NONE:
       return GST_VAAPI_RATECONTROL_NONE;
-#ifdef VA_RC_CQP
     case VA_RC_CQP:
       return GST_VAAPI_RATECONTROL_CQP;
-#endif
     case VA_RC_CBR:
       return GST_VAAPI_RATECONTROL_CBR;
     case VA_RC_VCM:
       return GST_VAAPI_RATECONTROL_VCM;
     case VA_RC_VBR:
       return GST_VAAPI_RATECONTROL_VBR;
-#ifdef VA_RC_VBR_CONSTRAINED
     case VA_RC_VBR_CONSTRAINED:
       return GST_VAAPI_RATECONTROL_VBR_CONSTRAINED;
-#endif
-#ifdef VA_RC_MB
+#if VA_CHECK_VERSION(0,39,1)
     case VA_RC_MB:
       return GST_VAAPI_RATECONTROL_MB;
 #endif
+#if VA_CHECK_VERSION(1,1,0)
+    case VA_RC_ICQ:
+      return GST_VAAPI_RATECONTROL_ICQ;
+#endif
+#if VA_CHECK_VERSION(1,3,0)
+    case VA_RC_QVBR:
+      return GST_VAAPI_RATECONTROL_QVBR;
+#endif
+
   }
   GST_ERROR ("unsupported VA-API Rate Control value %u", value);
   return GST_VAAPI_RATECONTROL_NONE;
@@ -828,3 +849,177 @@ to_GstVaapiScaleMethod (guint flags)
   }
   return method;
 }
+
+/* VPP: translate GstVideoOrientationMethod into VA mirror/rotation flags */
+void
+from_GstVideoOrientationMethod (guint value, guint * va_mirror,
+    guint * va_rotation)
+{
+  *va_mirror = 0;
+  *va_rotation = 0;
+
+  switch (value) {
+#if VA_CHECK_VERSION(1,1,0)
+    case GST_VIDEO_ORIENTATION_IDENTITY:
+      *va_mirror = VA_MIRROR_NONE;
+      *va_rotation = VA_ROTATION_NONE;
+      break;
+    case GST_VIDEO_ORIENTATION_HORIZ:
+      *va_mirror = VA_MIRROR_HORIZONTAL;
+      *va_rotation = VA_ROTATION_NONE;
+      break;
+    case GST_VIDEO_ORIENTATION_VERT:
+      *va_mirror = VA_MIRROR_VERTICAL;
+      *va_rotation = VA_ROTATION_NONE;
+      break;
+    case GST_VIDEO_ORIENTATION_90R:
+      *va_mirror = VA_MIRROR_NONE;
+      *va_rotation = VA_ROTATION_90;
+      break;
+    case GST_VIDEO_ORIENTATION_180:
+      *va_mirror = VA_MIRROR_NONE;
+      *va_rotation = VA_ROTATION_180;
+      break;
+    case GST_VIDEO_ORIENTATION_90L:
+      *va_mirror = VA_MIRROR_NONE;
+      *va_rotation = VA_ROTATION_270;
+      break;
+    case GST_VIDEO_ORIENTATION_UL_LR:
+      *va_mirror = VA_MIRROR_HORIZONTAL;
+      *va_rotation = VA_ROTATION_90;
+      break;
+    case GST_VIDEO_ORIENTATION_UR_LL:
+      *va_mirror = VA_MIRROR_VERTICAL;
+      *va_rotation = VA_ROTATION_90;
+      break;
+#endif
+    default:
+      break;
+  }
+}
+
+/**
+ * from_GstVaapiBufferMemoryType:
+ * @type: a #GstVaapiBufferMemoryType
+ *
+ * Returns: the VA's memory type symbol
+ **/
+guint
+from_GstVaapiBufferMemoryType (guint type)
+{
+  guint va_type;
+
+  switch (type) {
+#if VA_CHECK_VERSION(1,1,0)
+    case GST_VAAPI_BUFFER_MEMORY_TYPE_DMA_BUF2:
+      va_type = VA_SURFACE_ATTRIB_MEM_TYPE_DRM_PRIME_2;
+      break;
+#endif
+    case GST_VAAPI_BUFFER_MEMORY_TYPE_DMA_BUF:
+      va_type = VA_SURFACE_ATTRIB_MEM_TYPE_DRM_PRIME;
+      break;
+    case GST_VAAPI_BUFFER_MEMORY_TYPE_GEM_BUF:
+      va_type = VA_SURFACE_ATTRIB_MEM_TYPE_KERNEL_DRM;
+      break;
+    case GST_VAAPI_BUFFER_MEMORY_TYPE_V4L2:
+      va_type = VA_SURFACE_ATTRIB_MEM_TYPE_V4L2;
+      break;
+    case GST_VAAPI_BUFFER_MEMORY_TYPE_USER_PTR:
+      va_type = VA_SURFACE_ATTRIB_MEM_TYPE_USER_PTR;
+      break;
+    default:
+      va_type = 0;
+      break;
+  }
+  return va_type;
+}
+
+/**
+ * to_GstVaapiBufferMemoryType:
+ * @va_type: a VA's memory type symbol
+ *
+ * It will return the first "supported" memory type from @va_type bit
+ * flag.
+ *
+ * Returns: a #GstVaapiBufferMemoryType or 0 if unknown.
+ **/
+guint
+to_GstVaapiBufferMemoryType (guint va_type)
+{
+#if VA_CHECK_VERSION(1,1,0)
+  if ((va_type & VA_SURFACE_ATTRIB_MEM_TYPE_DRM_PRIME_2))
+    return GST_VAAPI_BUFFER_MEMORY_TYPE_DMA_BUF2;
+#endif
+  if ((va_type & VA_SURFACE_ATTRIB_MEM_TYPE_DRM_PRIME))
+    return GST_VAAPI_BUFFER_MEMORY_TYPE_DMA_BUF;
+  if ((va_type & VA_SURFACE_ATTRIB_MEM_TYPE_KERNEL_DRM))
+    return GST_VAAPI_BUFFER_MEMORY_TYPE_GEM_BUF;
+  if ((va_type & VA_SURFACE_ATTRIB_MEM_TYPE_V4L2))
+    return GST_VAAPI_BUFFER_MEMORY_TYPE_V4L2;
+  if ((va_type & VA_SURFACE_ATTRIB_MEM_TYPE_USER_PTR))
+    return GST_VAAPI_BUFFER_MEMORY_TYPE_USER_PTR;
+  return 0;
+}
+
+/**
+ * from_GstVideoColorimetry:
+ * @colorimetry: a #GstVideoColorimetry type
+ *
+ * VPP: maps the #GstVideoColorimetry type to the VAProcColorStandardType.  If
+ * @colorimetry is NULL or colorimetry->primaries are unknown, then returns
+ * VAProcColorStandardNone.  If there is no 1:1 correlation, then returns
+ * VAProcColorStandardExplicit.  Otherwise, the correlating
+ * VAProcColorStandardType is returned.
+ *
+ * Returns: a VAProcColorStandardType.
+ **/
+guint
+from_GstVideoColorimetry (const GstVideoColorimetry * const colorimetry)
+{
+#if VA_CHECK_VERSION(1,2,0)
+  if (!colorimetry
+      || colorimetry->primaries == GST_VIDEO_COLOR_PRIMARIES_UNKNOWN)
+    return VAProcColorStandardNone;
+  if (gst_video_colorimetry_matches (colorimetry, GST_VIDEO_COLORIMETRY_BT709))
+    return VAProcColorStandardBT709;
+  /* NOTE: VAProcColorStandardBT2020 in VAAPI is the same as
+   * GST_VIDEO_COLORIMETRY_BT2020_10 in gstreamer. */
+  if (gst_video_colorimetry_matches (colorimetry,
+          GST_VIDEO_COLORIMETRY_BT2020_10))
+    return VAProcColorStandardBT2020;
+  if (gst_video_colorimetry_matches (colorimetry, GST_VIDEO_COLORIMETRY_BT601))
+    return VAProcColorStandardBT601;
+  if (gst_video_colorimetry_matches (colorimetry,
+          GST_VIDEO_COLORIMETRY_SMPTE240M))
+    return VAProcColorStandardSMPTE240M;
+
+  return VAProcColorStandardExplicit;
+#else
+  return VAProcColorStandardNone;
+#endif
+}
+
+/**
+ * from_GstVideoColorRange:
+ * @value: a #GstVideoColorRange
+ *
+ * VPP: maps the #GstVideoColorRange to the VA value.
+ *
+ * Returns: the VA color range.
+ **/
+guint
+from_GstVideoColorRange (const GstVideoColorRange value)
+{
+#if VA_CHECK_VERSION(1,2,0)
+  switch (value) {
+    case GST_VIDEO_COLOR_RANGE_0_255:
+      return VA_SOURCE_RANGE_FULL;
+    case GST_VIDEO_COLOR_RANGE_16_235:
+      return VA_SOURCE_RANGE_REDUCED;
+    default:
+      return VA_SOURCE_RANGE_UNKNOWN;
+  }
+#else
+  return 0;
+#endif
+}
