@@ -24,6 +24,7 @@
 
 #include "gstcompat.h"
 #include <gst/vaapi/gstvaapidisplay.h>
+#include <gst/vaapi/gstvaapiprofilecaps.h>
 
 #include "gstvaapidecode.h"
 #include "gstvaapidecode_props.h"
@@ -77,12 +78,12 @@ static const char gst_vaapidecode_sink_caps_str[] =
     ;
 
 static const char gst_vaapidecode_src_caps_str[] =
-    GST_VAAPI_MAKE_SURFACE_CAPS ";"
+    GST_VAAPI_MAKE_SURFACE_CAPS "; "
+    GST_VIDEO_CAPS_MAKE_WITH_FEATURES(GST_CAPS_FEATURE_MEMORY_DMABUF, GST_VAAPI_FORMATS_ALL) " ;"
 #if (USE_GLX || USE_EGL)
-    GST_VAAPI_MAKE_GLTEXUPLOAD_CAPS ";"
+    GST_VAAPI_MAKE_GLTEXUPLOAD_CAPS "; "
 #endif
-    GST_VIDEO_CAPS_MAKE("{ NV12, I420, YV12, YUY2, UYVY, Y210, P010_10LE, AYUV, Y410 }") ";"
-    GST_VAAPI_MAKE_DMABUF_CAPS;
+    GST_VIDEO_CAPS_MAKE(GST_VAAPI_FORMATS_ALL);
 
 static GstStaticPadTemplate gst_vaapidecode_src_factory =
     GST_STATIC_PAD_TEMPLATE(
@@ -199,48 +200,112 @@ gst_vaapidecode_update_sink_caps (GstVaapiDecode * decode, GstCaps * caps)
 static gboolean
 gst_vaapidecode_ensure_allowed_srcpad_caps (GstVaapiDecode * decode)
 {
-  GstCaps *out_caps, *raw_caps;
+  GstVaapiDisplay *const display = GST_VAAPI_PLUGIN_BASE_DISPLAY (decode);
+  GstCaps *out_caps, *raw_caps, *va_caps, *dma_caps, *gltexup_caps, *base_caps;
+  GArray *formats;
+  gint min_width, min_height, max_width, max_height;
+  guint mem_types;
+  gboolean ret = FALSE;
 
   if (decode->allowed_srcpad_caps)
     return TRUE;
 
-  if (!GST_VAAPI_PLUGIN_BASE_DISPLAY (decode))
+  if (!display)
     return FALSE;
 
-  /* Create VA caps */
-  out_caps = gst_caps_from_string (GST_VAAPI_MAKE_SURFACE_CAPS);
-  if (!out_caps) {
-    GST_WARNING_OBJECT (decode, "failed to create VA/GL source caps");
+  if (!decode->decoder)
     return FALSE;
+
+  dma_caps = gltexup_caps = NULL;
+
+  formats = gst_vaapi_decoder_get_surface_attributes (decode->decoder,
+      &min_width, &min_height, &max_width, &max_height, &mem_types);
+  if (!formats)
+    return FALSE;
+
+  base_caps = gst_vaapi_video_format_new_template_caps_from_list (formats);
+  if (!base_caps)
+    goto bail;
+  gst_vaapi_caps_set_width_and_height_range (base_caps, min_width, min_height,
+      max_width, max_height);
+
+  {
+    GArray *img_formats = gst_vaapi_display_get_image_formats (display);
+    GstVideoFormat decoded_format =
+        GST_VIDEO_INFO_FORMAT (&decode->decoded_info);
+
+    if (!img_formats)
+      img_formats = g_array_ref (formats);
+
+    if (decoded_format != GST_VIDEO_FORMAT_UNKNOWN) {
+      guint decoded_chroma =
+          gst_vaapi_video_format_get_chroma_type (decoded_format);
+      GArray *new_img_formats =
+          g_array_new (FALSE, FALSE, sizeof (GstVideoFormat));
+      gint i;
+
+      for (i = 0; i < img_formats->len; i++) {
+        const GstVideoFormat fmt =
+            g_array_index (img_formats, GstVideoFormat, i);
+        if (gst_vaapi_video_format_get_chroma_type (fmt) == decoded_chroma)
+          g_array_append_val (new_img_formats, fmt);
+      }
+
+      if (new_img_formats->len == 0) {
+        g_clear_pointer (&new_img_formats, g_array_unref);
+      } else {
+        g_clear_pointer (&img_formats, g_array_unref);
+        img_formats = new_img_formats;
+      }
+    }
+
+    raw_caps = gst_vaapi_video_format_new_template_caps_from_list (img_formats);
+    gst_vaapi_caps_set_width_and_height_range (raw_caps, min_width, min_height,
+        max_width, max_height);
+    g_array_unref (img_formats);
+  }
+
+  va_caps = gst_caps_copy (base_caps);
+  gst_caps_set_features_simple (va_caps,
+      gst_caps_features_from_string (GST_CAPS_FEATURE_MEMORY_VAAPI_SURFACE));
+
+  if (GST_VAAPI_PLUGIN_BASE_SRC_PAD_CAN_DMABUF (decode)
+      && gst_vaapi_mem_type_supports (mem_types,
+          GST_VAAPI_BUFFER_MEMORY_TYPE_DMA_BUF)) {
+    dma_caps = gst_caps_copy (base_caps);
+    gst_caps_set_features_simple (va_caps,
+        gst_caps_features_from_string (GST_CAPS_FEATURE_MEMORY_DMABUF));
   }
 #if (USE_GLX || USE_EGL)
-  if (!GST_VAAPI_PLUGIN_BASE_SRC_PAD_CAN_DMABUF (decode) &&
-      gst_vaapi_display_has_opengl (GST_VAAPI_PLUGIN_BASE_DISPLAY (decode))) {
-    out_caps = gst_caps_make_writable (out_caps);
-    gst_caps_append (out_caps,
-        gst_caps_from_string (GST_VAAPI_MAKE_GLTEXUPLOAD_CAPS));
+  if (!GST_VAAPI_PLUGIN_BASE_SRC_PAD_CAN_DMABUF (decode)
+      && gst_vaapi_display_has_opengl (GST_VAAPI_PLUGIN_BASE_DISPLAY (decode))) {
+    gltexup_caps = gst_caps_from_string (GST_VAAPI_MAKE_GLTEXUPLOAD_CAPS);
+    if (gltexup_caps) {
+      gst_vaapi_caps_set_width_and_height_range (base_caps, min_width,
+          min_height, max_width, max_height);
+    }
   }
 #endif
-  out_caps = gst_caps_make_writable (out_caps);
-  gst_caps_append (out_caps, gst_caps_from_string (GST_VAAPI_MAKE_DMABUF_CAPS));
 
-  raw_caps = gst_vaapi_plugin_base_get_allowed_srcpad_raw_caps
-      (GST_VAAPI_PLUGIN_BASE (decode),
-      GST_VIDEO_INFO_FORMAT (&decode->decoded_info));
-  if (!raw_caps) {
-    gst_caps_unref (out_caps);
-    GST_WARNING_OBJECT (decode, "failed to create raw sink caps");
-    return FALSE;
-  }
-
-  out_caps = gst_caps_make_writable (out_caps);
-  gst_caps_append (out_caps, gst_caps_copy (raw_caps));
+  out_caps = va_caps;
+  if (dma_caps)
+    gst_caps_append (out_caps, dma_caps);
+  if (gltexup_caps)
+    gst_caps_append (out_caps, gltexup_caps);
+  gst_caps_append (out_caps, raw_caps);
   decode->allowed_srcpad_caps = out_caps;
 
   GST_INFO_OBJECT (decode, "allowed srcpad caps: %" GST_PTR_FORMAT,
       decode->allowed_srcpad_caps);
 
-  return TRUE;
+  ret = TRUE;
+
+bail:
+  if (formats)
+    g_array_unref (formats);
+  if (base_caps)
+    gst_caps_unref (base_caps);
+  return ret;
 }
 
 static GstCaps *
@@ -324,13 +389,8 @@ gst_vaapidecode_update_src_caps (GstVaapiDecode * decode)
     case GST_VAAPI_CAPS_FEATURE_DMABUF:
     case GST_VAAPI_CAPS_FEATURE_VAAPI_SURFACE:{
       GstStructure *structure = gst_caps_get_structure (state->caps, 0);
-
-      /* Remove chroma-site and colorimetry from src caps,
-       * which is unnecessary on downstream if using VASurface
-       */
-      gst_structure_remove_fields (structure, "chroma-site", "colorimetry",
-          NULL);
-
+      if (!structure)
+        break;
       feature_str = gst_vaapi_caps_feature_to_string (feature);
       features = gst_caps_features_new (feature_str, NULL);
       gst_caps_set_features (state->caps, 0, features);
@@ -379,15 +439,6 @@ gst_vaapidecode_update_src_caps (GstVaapiDecode * decode)
   return TRUE;
 }
 
-static void
-gst_vaapidecode_release (GstVaapiDecode * decode)
-{
-  g_mutex_lock (&decode->surface_ready_mutex);
-  g_cond_signal (&decode->surface_ready);
-  g_mutex_unlock (&decode->surface_ready_mutex);
-  gst_object_unref (decode);
-}
-
 /* check whether the decoded surface size has changed */
 static gboolean
 is_surface_resolution_changed (GstVaapiDecode * decode,
@@ -413,12 +464,21 @@ is_surface_resolution_changed (GstVaapiDecode * decode,
     surface_format = gst_vaapi_surface_get_format (surface);
 
     /* if the VA context delivers a currently unrecognized format
-     * (ICM3, e.g.), we can assume NV12 "safely" */
+     * (ICM3, e.g.), we can assume one according surface chroma
+     * type. If fail, then use NV12 "safely" */
     if (surface_format == GST_VIDEO_FORMAT_UNKNOWN
         || surface_format == GST_VIDEO_FORMAT_ENCODED)
+      surface_format =
+          gst_vaapi_video_format_from_chroma (gst_vaapi_surface_get_chroma_type
+          (surface));
+    if (surface_format == GST_VIDEO_FORMAT_UNKNOWN)
       surface_format = GST_VIDEO_FORMAT_NV12;
   }
 
+  /* reset allowed source caps since they are dependant of the decoded
+   * surface format */
+  gst_caps_replace (&decode->allowed_srcpad_caps, NULL);
+
   gst_video_info_set_format (vinfo, surface_format, surface_width,
       surface_height);
 
@@ -470,18 +530,22 @@ gst_vaapidecode_negotiate (GstVaapiDecode * decode)
   GST_DEBUG_OBJECT (decode, "input codec state changed: renegotiating");
 
   GST_VIDEO_DECODER_STREAM_LOCK (vdec);
-  if (!gst_vaapi_plugin_base_set_caps (plugin, decode->sinkpad_caps, NULL))
-    return FALSE;
   if (!gst_vaapidecode_update_src_caps (decode))
-    return FALSE;
+    goto caps_negotiation_failed;
   if (!gst_vaapi_plugin_base_set_caps (plugin, NULL, decode->srcpad_caps))
-    return FALSE;
+    goto caps_negotiation_failed;
   GST_VIDEO_DECODER_STREAM_UNLOCK (vdec);
 
   if (!gst_video_decoder_negotiate (vdec))
     return FALSE;
 
   return TRUE;
+
+caps_negotiation_failed:
+  {
+    GST_VIDEO_DECODER_STREAM_UNLOCK (vdec);
+    return FALSE;
+  }
 }
 
 static gboolean
@@ -491,7 +555,9 @@ is_src_allocator_dmabuf (GstVaapiDecode * decode)
 
   if (!GST_VAAPI_PLUGIN_BASE_SRC_PAD_CAN_DMABUF (plugin))
     return FALSE;
-  return gst_vaapi_is_dmabuf_allocator (plugin->srcpad_allocator);
+  return
+      gst_vaapi_is_dmabuf_allocator (GST_VAAPI_PLUGIN_BASE_SRC_PAD_ALLOCATOR
+      (plugin));
 }
 
 static GstFlowReturn
@@ -534,9 +600,6 @@ gst_vaapidecode_push_decoded_frame (GstVideoDecoder * vdec,
         return GST_FLOW_ERROR;
     }
 
-    gst_vaapi_surface_proxy_set_destroy_notify (proxy,
-        (GDestroyNotify) gst_vaapidecode_release, gst_object_ref (decode));
-
     if (is_src_allocator_dmabuf (decode)) {
       vaapi_params.proxy = gst_vaapi_surface_proxy_ref (proxy);
       params = (GstBufferPoolAcquireParams *) & vaapi_params;
@@ -588,9 +651,11 @@ gst_vaapidecode_push_decoded_frame (GstVideoDecoder * vdec,
       GstBuffer *sys_buf, *va_buf;
 
       va_buf = out_frame->output_buffer;
-      sys_buf = gst_buffer_new_allocate (plugin->other_srcpad_allocator,
-          GST_VIDEO_INFO_SIZE (&plugin->srcpad_info),
-          &plugin->other_allocator_params);
+      sys_buf =
+          gst_buffer_new_allocate (GST_VAAPI_PLUGIN_BASE_OTHER_ALLOCATOR
+          (plugin),
+          GST_VIDEO_INFO_SIZE (GST_VAAPI_PLUGIN_BASE_SRC_PAD_INFO (plugin)),
+          &GST_VAAPI_PLUGIN_BASE_OTHER_ALLOCATOR_PARAMS (plugin));
       if (!sys_buf)
         goto error_no_sys_buffer;
 
@@ -604,14 +669,6 @@ gst_vaapidecode_push_decoded_frame (GstVideoDecoder * vdec,
     }
   }
 
-  if (decode->in_segment.rate < 0.0
-      && !GST_VAAPI_PLUGIN_BASE_COPY_OUTPUT_FRAME (vdec)
-      && !GST_VIDEO_CODEC_FRAME_IS_SYNC_POINT (out_frame)) {
-    GST_TRACE_OBJECT (decode, "drop frame in reverse playback");
-    gst_video_decoder_release_frame (GST_VIDEO_DECODER (decode), out_frame);
-    return GST_FLOW_OK;
-  }
-
   ret = gst_video_decoder_finish_frame (vdec, out_frame);
   if (ret != GST_FLOW_OK)
     goto error_commit_buffer;
@@ -656,7 +713,7 @@ error_cannot_copy:
   }
 error_commit_buffer:
   {
-    GST_INFO_OBJECT (decode, "downstream element rejected the frame (%s [%d])",
+    GST_LOG_OBJECT (decode, "downstream element rejected the frame (%s [%d])",
         gst_flow_get_name (ret), ret);
     return ret;
   }
@@ -698,7 +755,6 @@ gst_vaapidecode_handle_frame (GstVideoDecoder * vdec,
 {
   GstVaapiDecode *const decode = GST_VAAPIDECODE (vdec);
   GstVaapiDecoderStatus status;
-  GstFlowReturn ret;
 
   if (!decode->input_state)
     goto not_negotiated;
@@ -706,20 +762,6 @@ gst_vaapidecode_handle_frame (GstVideoDecoder * vdec,
   /* Decode current frame */
   for (;;) {
     status = gst_vaapi_decoder_decode (decode->decoder, frame);
-    if (status == GST_VAAPI_DECODER_STATUS_ERROR_NO_SURFACE) {
-      /* Make sure that there are no decoded frames waiting in the
-         output queue. */
-      ret = gst_vaapidecode_push_all_decoded_frames (decode);
-      if (ret != GST_FLOW_OK)
-        goto error_push_all_decoded_frames;
-
-      g_mutex_lock (&decode->surface_ready_mutex);
-      if (gst_vaapi_decoder_check_status (decode->decoder) ==
-          GST_VAAPI_DECODER_STATUS_ERROR_NO_SURFACE)
-        g_cond_wait (&decode->surface_ready, &decode->surface_ready_mutex);
-      g_mutex_unlock (&decode->surface_ready_mutex);
-      continue;
-    }
     if (status != GST_VAAPI_DECODER_STATUS_SUCCESS)
       goto error_decode;
     break;
@@ -731,15 +773,12 @@ gst_vaapidecode_handle_frame (GstVideoDecoder * vdec,
   return gst_vaapidecode_push_all_decoded_frames (decode);
 
   /* ERRORS */
-error_push_all_decoded_frames:
-  {
-    GST_ERROR ("push loop error while decoding %d", ret);
-    gst_video_decoder_drop_frame (vdec, frame);
-    return ret;
-  }
 error_decode:
   {
-    GST_ERROR ("decode error %d", status);
+    GstFlowReturn ret = GST_FLOW_OK;
+
+    GST_WARNING_OBJECT (decode, "decode error %d", status);
+
     switch (status) {
       case GST_VAAPI_DECODER_STATUS_ERROR_UNSUPPORTED_CODEC:
       case GST_VAAPI_DECODER_STATUS_ERROR_UNSUPPORTED_PROFILE:
@@ -747,14 +786,12 @@ error_decode:
         ret = GST_FLOW_NOT_SUPPORTED;
         break;
       default:
-        ret = GST_FLOW_OK;
         GST_VIDEO_DECODER_ERROR (vdec, 1, STREAM, DECODE, ("Decoding error"),
             ("Decode error %d", status), ret);
-        GST_INFO ("requesting upstream a key unit");
+        GST_INFO_OBJECT (decode, "requesting upstream a key unit");
         gst_pad_push_event (GST_VIDEO_DECODER_SINK_PAD (decode),
             gst_video_event_new_upstream_force_key_unit (GST_CLOCK_TIME_NONE,
                 FALSE, 0));
-        ret = GST_FLOW_OK;
         break;
     }
     gst_video_decoder_drop_frame (vdec, frame);
@@ -763,9 +800,8 @@ error_decode:
 not_negotiated:
   {
     GST_ERROR_OBJECT (decode, "not negotiated");
-    ret = GST_FLOW_NOT_NEGOTIATED;
     gst_video_decoder_drop_frame (vdec, frame);
-    return ret;
+    return GST_FLOW_NOT_NEGOTIATED;
   }
 }
 
@@ -856,12 +892,6 @@ gst_vaapidecode_ensure_display (GstVaapiDecode * decode)
   return gst_vaapi_plugin_base_ensure_display (GST_VAAPI_PLUGIN_BASE (decode));
 }
 
-static inline guint
-gst_vaapi_codec_from_caps (GstCaps * caps)
-{
-  return gst_vaapi_profile_get_codec (gst_vaapi_profile_from_caps (caps));
-}
-
 static gboolean
 gst_vaapidecode_create (GstVaapiDecode * decode, GstCaps * caps)
 {
@@ -871,7 +901,7 @@ gst_vaapidecode_create (GstVaapiDecode * decode, GstCaps * caps)
     return FALSE;
   dpy = GST_VAAPI_PLUGIN_BASE_DISPLAY (decode);
 
-  switch (gst_vaapi_codec_from_caps (caps)) {
+  switch (gst_vaapi_get_codec_from_caps (caps)) {
     case GST_VAAPI_CODEC_MPEG2:
       decode->decoder = gst_vaapi_decoder_mpeg2_new (dpy, caps);
       break;
@@ -889,6 +919,9 @@ gst_vaapidecode_create (GstVaapiDecode * decode, GstCaps * caps)
         GstStructure *const structure = gst_caps_get_structure (caps, 0);
         const gchar *str = NULL;
 
+        if (!structure)
+          break;
+
         if ((str = gst_structure_get_string (structure, "alignment"))) {
           GstVaapiStreamAlignH264 alignment;
           if (g_strcmp0 (str, "au") == 0)
@@ -917,6 +950,9 @@ gst_vaapidecode_create (GstVaapiDecode * decode, GstCaps * caps)
         GstStructure *const structure = gst_caps_get_structure (caps, 0);
         const gchar *str = NULL;
 
+        if (!structure)
+          break;
+
         if ((str = gst_structure_get_string (structure, "alignment"))) {
           GstVaapiStreamAlignH265 alignment;
           if (g_strcmp0 (str, "au") == 0)
@@ -989,8 +1025,8 @@ gst_vaapidecode_destroy (GstVaapiDecode * decode)
   gst_vaapidecode_purge (decode);
 
   gst_vaapi_decoder_replace (&decode->decoder, NULL);
-
-  gst_vaapidecode_release (gst_object_ref (decode));
+  /* srcpad caps are decoder's context dependant */
+  gst_caps_replace (&decode->allowed_srcpad_caps, NULL);
 }
 
 static gboolean
@@ -1018,11 +1054,6 @@ gst_vaapidecode_reset (GstVaapiDecode * decode, GstCaps * caps,
 static void
 gst_vaapidecode_finalize (GObject * object)
 {
-  GstVaapiDecode *const decode = GST_VAAPIDECODE (object);
-
-  g_cond_clear (&decode->surface_ready);
-  g_mutex_clear (&decode->surface_ready_mutex);
-
   gst_vaapi_plugin_base_finalize (GST_VAAPI_PLUGIN_BASE (object));
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
@@ -1203,30 +1234,33 @@ is_svc_profile (GstVaapiProfile profile)
       || profile == GST_VAAPI_PROFILE_H264_SCALABLE_HIGH;
 }
 
-
-static GstCaps *
-add_h264_profile_in_caps (GstCaps * caps, const gchar * profile_name)
+static void
+find_mvc_and_svc (GArray * profiles, gboolean * have_mvc, gboolean * have_svc)
 {
-  GstCaps *caps_new =
-      gst_caps_new_simple ("video/x-h264", "profile", G_TYPE_STRING,
-      profile_name, NULL);
-  return gst_caps_merge (caps_new, caps);
+  guint i;
+
+  for (i = 0; i < profiles->len; i++) {
+    const GstVaapiProfile profile =
+        g_array_index (profiles, GstVaapiProfile, i);
+
+    *have_mvc |= is_mvc_profile (profile);
+    *have_svc |= is_svc_profile (profile);
+  }
 }
 
 static gboolean
 gst_vaapidecode_ensure_allowed_sinkpad_caps (GstVaapiDecode * decode)
 {
   GstCaps *caps, *allowed_sinkpad_caps;
+  GstPad *const sinkpad = GST_VIDEO_DECODER_SINK_PAD (decode);
   GArray *profiles;
+  GstVaapiDisplay *const display = GST_VAAPI_PLUGIN_BASE_DISPLAY (decode);
   guint i;
   gboolean base_only = FALSE;
-  gboolean have_high = FALSE;
   gboolean have_mvc = FALSE;
   gboolean have_svc = FALSE;
 
-  profiles =
-      gst_vaapi_display_get_decode_profiles (GST_VAAPI_PLUGIN_BASE_DISPLAY
-      (decode));
+  profiles = gst_vaapi_display_get_decode_profiles (display);
   if (!profiles)
     goto error_no_profiles;
 
@@ -1238,6 +1272,8 @@ gst_vaapidecode_ensure_allowed_sinkpad_caps (GstVaapiDecode * decode)
     g_object_get (decode, "base-only", &base_only, NULL);
   }
 
+  find_mvc_and_svc (profiles, &have_mvc, &have_svc);
+
   for (i = 0; i < profiles->len; i++) {
     const GstVaapiProfile profile =
         g_array_index (profiles, GstVaapiProfile, i);
@@ -1253,54 +1289,82 @@ gst_vaapidecode_ensure_allowed_sinkpad_caps (GstVaapiDecode * decode)
     if (!caps)
       continue;
     structure = gst_caps_get_structure (caps, 0);
+    if (!structure)
+      continue;
 
     profile_name = gst_vaapi_profile_get_name (profile);
-    if (profile_name)
+    if (!profile_name)
+      goto merge_caps;
+
+    /* Add all according -intra profile for HEVC */
+    if (profile == GST_VAAPI_PROFILE_H265_MAIN
+        || profile == GST_VAAPI_PROFILE_H265_MAIN10
+        || profile == GST_VAAPI_PROFILE_H265_MAIN_422_10
+        || profile == GST_VAAPI_PROFILE_H265_MAIN_444
+        || profile == GST_VAAPI_PROFILE_H265_MAIN_444_10
+        || profile == GST_VAAPI_PROFILE_H265_MAIN12) {
+      gchar *profiles[3], *intra_name;
+
+      intra_name = g_strdup_printf ("%s-intra", profile_name);
+
+      profiles[0] = (gchar *) profile_name;
+      profiles[1] = intra_name;
+      profiles[2] = NULL;
+
+      gst_vaapi_structure_set_profiles (structure, profiles);
+      g_free (intra_name);
+
+    } else if (profile == GST_VAAPI_PROFILE_H264_CONSTRAINED_BASELINE) {
+      /* XXX: artificially adding baseline if constrained_baseline is
+       * available. */
+      gchar *profiles[] = { (gchar *) profile_name, "baseline", NULL };
+
+      gst_vaapi_structure_set_profiles (structure, profiles);
+    } else if (profile == GST_VAAPI_PROFILE_H264_HIGH) {
+      gchar *profiles[11] = { (gchar *) profile_name, "progressive-high",
+        "constrained-high"
+      };
+      gint i = 3;
+
+      if (base_only && !have_mvc) {
+        GST_DEBUG ("base_only: force adding MVC profiles in caps");
+
+        profiles[i++] = "multiview-high";
+        profiles[i++] = "stereo-high";
+      }
+
+      if (base_only && !have_svc) {
+        GST_DEBUG ("base_only: force adding SVC profiles in caps");
+
+        profiles[i++] = "scalable-constrained-baseline";
+        profiles[i++] = "scalable-baseline";
+        profiles[i++] = "scalable-high-intra";
+        profiles[i++] = "scalable-constrained-high";
+        profiles[i++] = "scalable-high";
+      }
+
+      profiles[i++] = NULL;
+
+      gst_vaapi_structure_set_profiles (structure, profiles);
+    } else {
       gst_structure_set (structure, "profile", G_TYPE_STRING,
           profile_name, NULL);
+    }
 
+  merge_caps:
+    gst_vaapi_profile_caps_append_decoder (display, profile, structure);
     allowed_sinkpad_caps = gst_caps_merge (allowed_sinkpad_caps, caps);
-    have_mvc |= is_mvc_profile (profile);
-    have_svc |= is_svc_profile (profile);
-    have_high |= profile == GST_VAAPI_PROFILE_H264_HIGH;
   }
 
-  if (have_high) {
-    allowed_sinkpad_caps =
-        add_h264_profile_in_caps (allowed_sinkpad_caps, "progressive-high");
-    allowed_sinkpad_caps =
-        add_h264_profile_in_caps (allowed_sinkpad_caps, "constrained-high");
-  }
-
-  if (base_only && (!have_mvc || !have_svc) && have_high) {
-    if (!have_mvc) {
-      GST_DEBUG ("base_only: force adding MVC profiles in caps");
-
-      allowed_sinkpad_caps =
-          add_h264_profile_in_caps (allowed_sinkpad_caps, "multiview-high");
-      allowed_sinkpad_caps =
-          add_h264_profile_in_caps (allowed_sinkpad_caps, "stereo-high");
-    }
-
-    if (!have_svc) {
-      GST_DEBUG ("base_only: force adding SVC profiles in caps");
-
-      allowed_sinkpad_caps =
-          add_h264_profile_in_caps (allowed_sinkpad_caps,
-          "scalable-constrained-baseline");
-      allowed_sinkpad_caps =
-          add_h264_profile_in_caps (allowed_sinkpad_caps, "scalable-baseline");
-      allowed_sinkpad_caps =
-          add_h264_profile_in_caps (allowed_sinkpad_caps,
-          "scalable-high-intra");
-      allowed_sinkpad_caps =
-          add_h264_profile_in_caps (allowed_sinkpad_caps,
-          "scalable-constrained-high");
-      allowed_sinkpad_caps =
-          add_h264_profile_in_caps (allowed_sinkpad_caps, "scalable-high");
-    }
-  }
-  decode->allowed_sinkpad_caps = gst_caps_simplify (allowed_sinkpad_caps);
+  caps = gst_pad_get_pad_template_caps (sinkpad);
+  decode->allowed_sinkpad_caps =
+      gst_caps_intersect (allowed_sinkpad_caps, caps);
+  gst_caps_unref (caps);
+  gst_caps_unref (allowed_sinkpad_caps);
+  decode->allowed_sinkpad_caps =
+      gst_caps_simplify (decode->allowed_sinkpad_caps);
+  GST_DEBUG_OBJECT (decode, "allowed sink caps %" GST_PTR_FORMAT,
+      decode->allowed_sinkpad_caps);
 
   g_array_unref (profiles);
   return TRUE;
@@ -1375,10 +1439,9 @@ gst_vaapidecode_src_query (GstVideoDecoder * vdec, GstQuery * query)
   switch (GST_QUERY_TYPE (query)) {
     case GST_QUERY_CAPS:{
       GstCaps *caps, *filter = NULL;
-      GstPad *pad = GST_VIDEO_DECODER_SRC_PAD (vdec);
 
       gst_query_parse_caps (query, &filter);
-      caps = gst_pad_get_pad_template_caps (pad);
+      caps = gst_vaapidecode_get_allowed_srcpad_caps (GST_VAAPIDECODE (vdec));
 
       if (filter) {
         GstCaps *tmp = caps;
@@ -1404,24 +1467,19 @@ gst_vaapidecode_src_query (GstVideoDecoder * vdec, GstQuery * query)
 }
 
 static gboolean
-gst_vaapidecode_sink_event (GstVideoDecoder * vdec, GstEvent * event)
+gst_vaapidecode_transform_meta (GstVideoDecoder *
+    vdec, GstVideoCodecFrame * frame, GstMeta * meta)
 {
-  GstVaapiDecode *const decode = GST_VAAPIDECODE (vdec);
+  const GstMetaInfo *info = meta->info;
 
-  switch (GST_EVENT_TYPE (event)) {
-    case GST_EVENT_SEGMENT:
-    {
-      /* Keep segment event to refer to rate so that
-       * vaapidecode can handle reverse playback
-       */
-      gst_event_copy_segment (event, &decode->in_segment);
-      break;
-    }
-    default:
-      break;
-  }
+  if (GST_VIDEO_DECODER_CLASS (parent_class)->transform_meta (vdec, frame,
+          meta))
+    return TRUE;
 
-  return GST_VIDEO_DECODER_CLASS (parent_class)->sink_event (vdec, event);
+  if (!g_strcmp0 (g_type_name (info->type), "GstVideoRegionOfInterestMeta"))
+    return TRUE;
+
+  return FALSE;
 }
 
 static void
@@ -1459,7 +1517,8 @@ gst_vaapidecode_class_init (GstVaapiDecodeClass * klass)
   vdec_class->src_query = GST_DEBUG_FUNCPTR (gst_vaapidecode_src_query);
   vdec_class->sink_query = GST_DEBUG_FUNCPTR (gst_vaapidecode_sink_query);
   vdec_class->getcaps = GST_DEBUG_FUNCPTR (gst_vaapidecode_sink_getcaps);
-  vdec_class->sink_event = GST_DEBUG_FUNCPTR (gst_vaapidecode_sink_event);
+  vdec_class->transform_meta =
+      GST_DEBUG_FUNCPTR (gst_vaapidecode_transform_meta);
 
   map = (GstVaapiDecoderMap *) g_type_get_qdata (G_OBJECT_CLASS_TYPE (klass),
       GST_VAAPI_DECODE_PARAMS_QDATA);
@@ -1507,9 +1566,6 @@ gst_vaapidecode_init (GstVaapiDecode * decode)
 
   gst_vaapi_plugin_base_init (GST_VAAPI_PLUGIN_BASE (decode), GST_CAT_DEFAULT);
 
-  g_mutex_init (&decode->surface_ready_mutex);
-  g_cond_init (&decode->surface_ready);
-
   gst_video_decoder_set_packetized (vdec, FALSE);
 }
 
